/* -*- buffer-read-only: t -*- vi: set ro:
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
*/
#define MF_SOURCE_NAME  MF_SOURCE_BI_IO
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include "mailfromd.h"
#include "prog.h"
#include "builtin.h"

#line 1 "bi_io.m4"
/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#define NSTREAMS 1024

struct io_stream {
	char *name;
	mu_locker_t lock;
	int fd[2];
	pid_t pid;
	char *buf;
	size_t bufsize;
};

#define IFD(s) ((s).fd[0])
#define OFD(s) ((s).fd[1] == -1 ? (s).fd[0] : (s).fd[1])

static void
flush_stream(struct io_stream *str)
{
	/*FIXME*/
}

static void
close_stream(struct io_stream *str)
{
	if (OFD(*str) == -1)
		return;
	flush_stream(str);
	close(OFD(*str));
	if (IFD(*str) != -1)
		close(IFD(*str));
	if (str->pid) {
		int status;
		waitpid(str->pid, &status, 0);
	}
	str->fd[0] = -1;
	str->fd[1] = -1;
	str->pid = 0;
}

static int
read_stream_line(struct io_stream *str)
{
	int fd = IFD(*str);
	size_t i = 0;
	int rc;

	for (i = 0; ; i++) {
		if (str->bufsize == i) {
			if (str->bufsize == 0) 
				str->bufsize = 16;
			str->buf = x2nrealloc(str->buf, &str->bufsize,
					      sizeof str->buf[1]);
		}
		rc = read(fd, str->buf + i, 1);
		if (rc == -1)
			return -1;
		if (str->buf[i] == '\n') {
			str->buf[i] = 0;
			break;
		}
	}
	return 0;
}

#define REDIRECT_STDIN_P(f) ((f) & (O_WRONLY|O_RDWR))
#define REDIRECT_STDOUT_P(f) ((f) == O_RDONLY || ((f) & (O_RDONLY|O_RDWR)))

static int
open_program_stream(struct io_stream *str, const char *cmdline, int flags)
{
	int rightp[2], leftp[2];
	int i;
	int rc = 0;
	pid_t pid;
	int argc;
	char **argv;
	
	if (REDIRECT_STDIN_P(flags))
		pipe(leftp);
	if (REDIRECT_STDOUT_P(flags))
		pipe(rightp);

	switch (pid = fork()) {
		/* The child branch.  */
	case 0:
		/* attach the pipes */

		/* Right-end */
		if (REDIRECT_STDOUT_P(flags)) {
			if (rightp[1] != 1) {
				close(1);
				dup2(rightp[1], 1);
			}
			close(rightp[0]);
		}

		/* Left-end */
		if (REDIRECT_STDIN_P(flags)) {
			if (leftp[0] != 0) {
				close(0);
				dup2(leftp[0], 0);
			}
			close(leftp[1]);
		}
      
		/* FIXME: Error output */
#if 0
		if (errfile) {
			i = open (errfile, O_CREAT|O_WRONLY|O_APPEND, 0644);
			if (i > 0 && i != 2) {
				dup2(i, 2);
				close(i);
			}
		}
#endif	
		/* Close unneded descripitors */
		for (i = getmaxfd(); i > 2; i--)
			close(i);

		debug1(10, "running %s", cmdline);
		rc = mu_argcv_get(cmdline, "", NULL, &argc, &argv);
		if (rc) {
			mu_error(_("Cannot parse command line %s: %s"),
				 cmdline, mu_strerror(rc));
			exit(127);
		}			
		execvp(argv[0], argv);
		mu_error(_("Cannot run %s: %s"),
			 cmdline, mu_strerror(rc));
		exit(127);
		/********************/

		/* Parent branches: */
	case -1:
		/* Fork has failed */
		/* Restore things */
		rc = errno;
		if (REDIRECT_STDOUT_P(flags)) {
			close(rightp[0]);
			close(rightp[1]);
		}
		if (REDIRECT_STDIN_P(flags)) {
			close(leftp[0]);
			close(leftp[1]);
		}
		break;
		
	default:
		str->pid = pid;
		if (REDIRECT_STDOUT_P(flags)) {
			str->fd[0] = rightp[0];
			close(rightp[1]);
		} else
			str->fd[0] = -1;

		if (REDIRECT_STDIN_P(flags)) {
			str->fd[1] = leftp[1];
			close(leftp[0]);
		} else
			str->fd[1] = -1;
	}
	return rc;
}
	
static int
open_file_stream(struct io_stream *str, const char *file, int flags)
{
	str->fd[0] = open(file, flags, 0644); /* FIXME: mode? */
	if (str->fd[0] == -1)
		return errno;
	return 0;
}

static void *
alloc_streams()
{
	struct io_stream *p, *stab = xcalloc(NSTREAMS, sizeof *stab);
	for (p = stab; p < stab + NSTREAMS; p++) 
		p->fd[0] = p->fd[1] = -1;
	return stab;
}

static void
destroy_streams(void *data)
{
	struct io_stream *stab = data;
	struct io_stream *p;
	for (p = stab; p < stab + NSTREAMS; p++) {
		close_stream(p);
		free(p->buf);
	}
	free(stab);
}
		

#line 214

#line 214
static int IO_id;
#line 214 "bi_io.m4"



void
#line 217
bi_open(eval_environ_t env)
#line 217

#line 217 "bi_io.m4"
{
#line 217
	
#line 217

#line 217

#line 217
char * name;
#line 217
        
#line 217
get_string_arg(env, 0, &name);
#line 217
        
#line 217

#line 217
        adjust_stack(env, 1);
#line 217

#line 217
	if (prog_trace_option & BUILTIN_MASK_io)
#line 217
		prog_trace(env, "open %s",name);;
#line 217

{
	int i, rc;
	int flags = 0;
	int (*opf)(struct io_stream *, const char *, int) = open_file_stream;
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	
	for (i = 0; i < NSTREAMS; i++) {
		if (iotab[i].fd[0] == -1) 
			break;
	}
		if (!(i < NSTREAMS))
#line 228
			env_throw_bi(env, mfe_failure, "open", _("No more files available"))
#line 230
;

	debug1(10, "opening stream %s", name);
	iotab[i].name = name;
	if (*name == '>') {
		flags |= O_RDWR|O_CREAT;
		name++;
		if (*name == '>') {
			flags |= O_APPEND;
			name++;
		} else
			flags |= O_TRUNC;
	} else if (*name == '|') {
		opf = open_program_stream;
		flags = O_WRONLY;
		name++;
		if (*name == '&') {
			flags = O_RDWR;
			name++;
		} 
	} else
		flags = O_RDWR;
	
	for (;*name && c_isspace(*name); name++)
		;
	
	rc = opf(&iotab[i], name, flags);
	
		if (!(rc == 0))
#line 258
			env_throw_bi(env, mfe_failure, "open", _("Cannot open stream %s: %s"),name,mu_strerror(rc))
#line 261
;
	debug2(10, "open(%s) = %d", name, i);
	
#line 263
do {
#line 263
  push(env, (STKVAL) (i));
#line 263
  return;
#line 263
} while (0);
}
}

void
#line 267
bi_close(eval_environ_t env)
#line 267

#line 267 "bi_io.m4"
{
#line 267
	
#line 267

#line 267

#line 267
long  fd;
#line 267
        
#line 267
get_numeric_arg(env, 0, &fd);
#line 267
        
#line 267

#line 267
        adjust_stack(env, 1);
#line 267

#line 267
	if (prog_trace_option & BUILTIN_MASK_io)
#line 267
		prog_trace(env, "close %lu",fd);;
#line 267

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);

		if (!(fd >= 0 && fd < NSTREAMS))
#line 271
			env_throw_bi(env, mfe_range, "close", _("Invalid file descriptor"))
#line 273
;
	close_stream(&iotab[fd]);
}
}	

void
#line 278
bi_write(eval_environ_t env)
#line 278

#line 278 "bi_io.m4"
{
#line 278
	
#line 278

#line 278
long __bi_argcnt;
#line 278
long  fd;
#line 278
        char * str;
#line 278
        long  n;
#line 278
        
#line 278
get_numeric_arg(env, 1, &fd);
#line 278
        get_string_arg(env, 2, &str);
#line 278
        get_numeric_arg(env, 3, &n);
#line 278
        
#line 278
get_numeric_arg(env, 0, &__bi_argcnt);
#line 278
        adjust_stack(env, __bi_argcnt + 1);
#line 278

#line 278
	if (prog_trace_option & BUILTIN_MASK_io)
#line 278
		prog_trace(env, "write %lu %s %lu",fd, str, ((__bi_argcnt > 2) ? n : 0));;
#line 278

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;
	
	debug2(10, "writing %s to %d", str, fd);
		if (!(fd >= 0 && fd < NSTREAMS && OFD(iotab[fd])))
#line 284
			env_throw_bi(env, mfe_range, "write", _("Invalid file descriptor"))
#line 286
;
	rc = write(OFD(iotab[fd]), str, ((__bi_argcnt > 2) ? n : strlen (str)));
		if (!(n == rc))
#line 288
			env_throw_bi(env, mfe_io, "write", _("Write error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 291
;
}
}

void
#line 295
bi_getline(eval_environ_t env)
#line 295

#line 295 "bi_io.m4"
{
#line 295
	
#line 295

#line 295

#line 295
long  fd;
#line 295
        
#line 295
get_numeric_arg(env, 0, &fd);
#line 295
        
#line 295

#line 295
        adjust_stack(env, 1);
#line 295

#line 295
	if (prog_trace_option & BUILTIN_MASK_io)
#line 295
		prog_trace(env, "getline %lu",fd);;
#line 295

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;

		if (!(fd >= 0 && fd < NSTREAMS && IFD(iotab[fd])))
#line 300
			env_throw_bi(env, mfe_range, "getline", _("Invalid file descriptor"))
#line 302
;
	rc = read_stream_line(&iotab[fd]);
		if (!(rc == 0))
#line 304
			env_throw_bi(env, mfe_io, "getline", _("Read error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 307
;
	do {pushs(env, iotab[fd].buf);
#line 308
return;
#line 308
} while (0);
}	
}


#line 312
void
#line 312
io_init_builtin()
#line 312
{
#line 312
	#line 214 "bi_io.m4"
IO_id = builtin_priv_register(alloc_streams, destroy_streams);
#line 217 "bi_io.m4"
va_builtin_install_ex("open", bi_open, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 267 "bi_io.m4"
va_builtin_install_ex("close", bi_close, 0, dtype_unspecified, 1, 0, 0|0, dtype_number);
#line 278 "bi_io.m4"
va_builtin_install_ex("write", bi_write, 0, dtype_unspecified, 3, 1, 0|0, dtype_number, dtype_string, dtype_number);
#line 295 "bi_io.m4"
va_builtin_install_ex("getline", bi_getline, 0, dtype_string, 1, 0, 0|0, dtype_number);

#line 312 "bi_io.m4"
		
#line 312
}
#line 594 "../mfd/snarf.m4"

#line 594

